/*****************************************************************
 * Copyright (c) ParaSoft Corporation 1991.  All rights reserved.*
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft       *
 * The copyright notice above does not evidence any              *
 * actual or intended publication of such source code.           *
 *****************************************************************/

#include "extern.h"

int se_in_decl_initexpr_parse = 0;

decl_check_entry (entry)
T_Entry *entry;
{
	if (IS_has_initexpr (entry)) {
		if (IS_argument (entry)
		||  IS_result (entry)
		|| (scoping_unit [state_level] != UNIT_BLOCK_DATA && IS_named_common (entry))
		||  IS_blank_common (entry)
		||  IS_external (entry)
		||  IS_intrinsic (entry)
		||  IS_automatic (entry)) {
			sprintf (serr
			, "initialization expression not allowed for '%s'"
			, entry->orig_name);
			fatal (serr);  /* E_5_2_46 */
		}
	}

	if (IS_parameter (entry)) {
		if (IS_argument (entry)
		||  IS_function (entry)
		||  IS_in_common (entry)) {
			sprintf (serr
			, "PARAMETER attribute not allowed for '%s'"
			, entry->orig_name);
			fatal (serr);  /* E_5_3_15 */
		}
	}

	if (IS_save (entry)) {
		if (IS_argument (entry)
		||  IS_procedure (entry)
		||  IS_result (entry)
		||  IS_automatic (entry)) {
			sprintf (serr
			, "SAVE attribute not allowed for '%s'"
			, entry->orig_name);
			fatal (serr);  /* E_5_3_20 */
		}
	}

	if (IS_external (entry) && IS_intrinsic (entry)) {
		sprintf (serr
		, "'%s' has both EXTERNAL and INTERNAL attributes"
		, entry->orig_name);
		fatal (serr);  /* E_5_3_22 */
	}
}

decl_initexpr_parse ()  /* initializes the variables that have an initialization Expr */
{
	T_Entry *entry;
	Node *expr, *stmt, *var;

	se_in_decl_initexpr_parse = 1;

	for_TABLE_LOOP (entry)
		if (IS_has_initexpr (entry) && (expr = ENTRY_INITEXPR (entry)) != NN) {
			var = make_Variable (make_Name (entry->orig_name));
			stmt = make_AssignementStmt (var, expr->OptInitializationExpr.Expr);
			stmt = parse_pass (stmt);
			block_add_stmt_init (stmt);
		}

	se_in_decl_initexpr_parse = 0;
}

/******************************* C H E C K **********************************/

static attr_more_than_once (attr)
char *attr;
{
	sprintf (serr, "attribute %s appears more than once", attr);  /* E_5_2_39 */
	fatal (serr);
}

Node *check_AttrSpec (root)
Node *root;
{
	Node *Nname = root->AttrSpec.Name;
	Node *sublist = root->AttrSpec.SubscriptList;
	char *name = U_strupper (Nname->com.str);

	if (!strncmp (name, "PARAMETER", 9)) {
		if (se_AttrSpec_param)
			attr_more_than_once ("PARAMETER");

		se_AttrSpec_param = 1;
	}
	else if (!strncmp (name, "PUBLIC", 6)) {
		if (se_AttrSpec_public)
			attr_more_than_once ("PUBLIC");

		se_AttrSpec_public = 1;
	}
	
	else if (!strncmp (name, "PRIVATE", 7)) {
		if (se_AttrSpec_private)
			attr_more_than_once ("PRIVATE");

		se_AttrSpec_private = 1;
	}
	else if (!strncmp (name, "ALLOCATABLE", 11)) {
		if (se_AttrSpec_allocatable)
			attr_more_than_once ("ALLOCATABLE");

		se_AttrSpec_allocatable = 1;
	}
	else if (!strncmp (name, "DIMENSION", 9) || !strncmp (name, "ARRAY", 5)) {
		if (se_AttrSpec_list != NN)
			attr_more_than_once ("DIMENSION/ARRAY");

		se_AttrSpec_list = sublist;

		if (name [0] == 'A')
			stand ("attribute ARRAY");
	}
	else if (!strncmp (name, "DATA", 4)) {
		if (se_AttrSpec_data)
			attr_more_than_once ("DATA");

		se_AttrSpec_data = 1;

		stand ("attribute DATA");
	}
	else if (!strncmp (name, "EXTERNAL", 8)) {
		if (se_AttrSpec_external)
			attr_more_than_once ("EXTERNAL");

		se_AttrSpec_external = 1;
	}
	else if (!strncmp (name, "INTENT", 6)) {
		if (se_AttrSpec_intent)
			attr_more_than_once ("INTENT");

		if (sublist != NN && sublist->com.type == NAME)
			se_AttrSpec_intent = intent_type (sublist->com.str);
		else
			se_AttrSpec_intent = 1;
	}
	else if (!strncmp (name, "INTRINSIC", 9)) {
		if (se_AttrSpec_intrinsic)
			attr_more_than_once ("INTRINSIC");

		se_AttrSpec_intrinsic = 1;
	}
	else if (!strncmp (name, "OPTIONAL", 8)) {
		if (se_AttrSpec_opt)
			attr_more_than_once ("OPTIONAL");

		se_AttrSpec_opt = 1;
	}
	else if (!strncmp (name, "POINTER", 7)) {
		if (se_AttrSpec_pointer)
			attr_more_than_once ("POINTER");

		se_AttrSpec_pointer = 1;
	}
	else if (!strncmp (name, "SAVE", 4)) {
		if (se_AttrSpec_save)
			attr_more_than_once ("SAVE");

		se_AttrSpec_save = 1;
	}
	else if (!strncmp (name, "TARGET", 6)) {
		if (se_AttrSpec_target)
			attr_more_than_once ("TARGET");

		se_AttrSpec_target = 1;
	}

	return root;
}

/******************************* W R I T E **********************************/

write_TypeDeclarationStmt (node)
Node *node;
{
	node->TypeDeclarationStmt.Comma = NN;
	node->TypeDeclarationStmt.AttrSpecList = NN;
	node->TypeDeclarationStmt.Colon = NN;
	node->TypeDeclarationStmt.Colon1 = NN;

	write_pass_childs (node);
}

write_EntityDecl (node)
Node *node;
{
	T_Entry *entry;

	if ((entry = lookup_loc (node->com.str)) != TN) {
/* ##BEGIN##, 02/23/1992, ZEKI */
/* eliminate the writing of automatic temporary which is output of intrinsic
  which I set off automatic before comming here.                            */
             
               /** BOZKUS **/
               node = change_EntityDecl(node);
              /** END **/

        /*     if (IS_automatic (entry) || IS_alloc_temp (entry)) */
               if (IS_automatic (entry) && IS_alloc_temp (entry))
/* ##END## */		auto_write_decl (node);
		else if (IS_assumed_shape (entry))
			assumed_write_decl (node);			
		else
			write_pass_childs (node);
	}
	else {
		write_pass_childs (node);
	}
}

/****************************************************************/
typedef struct Member {
	char *type, *var;
} Member;

static U_DEFINE_STACK (list, Member*, 10)

decl_intrin_add (type, var)
char *type, *var;
{
	Member *member;
	int i;

	for (i = 0; i < list_top; i++)
		if (!strcmp (list [i]->var, var))
			return;

	member = (Member*) U_malloc (sizeof (Member));
	member->type = type;
	member->var = var;

	U_stack_push_list (member);
}

decl_intrin_end_write ()
{
	U_stack_clear_list ();
}

decl_intrin_decl ()
{
	int i;

	for (i = 0; i < list_top; i++) {
		otab ();
		ostr (list [i]->type);
		onam (list [i]->var);
		odon ();
	}
}

write_IntrinsicStmt (node)  /* to not declare array intrinsics */
Node *node;
{
	Node *namelist = node->IntrinsicStmt.NameList;
	Node *Nname, *temp;
	char *name;
	T_Entry *entry;
	int none;

	none = 1;
	for (LIST_LOOP (Nname, namelist, temp))
		if ((entry = lookup_intrin (Nname->com.str)) == TN
		|| !IS_array_intrinsic (entry)) {
			none = 0;
			break;
		}

	if (none)
		return;

	oINTRINSIC ();

	none = 0;
	for (LIST_LOOP (Nname, namelist, temp))
		if ((entry = lookup_intrin (name = Nname->com.str)) == TN
		|| !IS_array_intrinsic (entry)) {
			if (none)
				oCOMMA ();
			none = 1;

			onam (name);
		}
}
